import abc
import time
from collections import deque
from concurrent.futures import Future, ThreadPoolExecutor
from datetime import datetime

import loggi
import scrapetools
from noiftimer import Timer
from pathier import Pathier, Pathish
from printbuddies import ColorMap, Progress, TimerColumn
from rich.console import Console
from rich.progress import ProgressColumn, TaskID
from seleniumuser.seleniumuser import User
from typing_extensions import Any, Callable, Sequence, override

from .core import ChoresMixin, ParserMixin, ScraperMetricsMixin
from .models import Url
from .requests import Response, SeleniumResponse, request

root = Pathier(__file__).parent
color_map = ColorMap()
console = Console(style="pink1")


class ThreadManager:
    """
    Manages workers when used with `concurrent.futures.ThreadPoolExecutor`.

    Pass the returned future from `concurrent.futures.ThreadPoolExecutor.submit()` to `add_future()`.

    e.g.
    >>> list_of_functions = [some_function for i in range(1000)]
    >>> thread_manager = ThreadManager()
    >>> with concurrent.futures.ThreadPoolExecutor() as executor:
    >>>   while list_of_functions or thread_manager.unfinished_workers:
    >>>     if thread_manager.open_slots:
    >>>       future = executor.submit(list_of_functions.pop())
    >>>       thread_manager.add_future(future)
    """

    def __init__(self, max_workers: int):
        self.max_workers = max_workers
        self.workers: list[Future[Any]] = []

    @property
    def cancelled_workers(self) -> list[Future[Any]]:
        """Returns a list of cancelled workers."""
        return [worker for worker in self.workers if worker.cancelled()]

    @property
    def completed_workers(self) -> list[Future[Any]]:
        """Returns a list of completed workers."""
        return [
            worker
            for worker in self.workers
            if worker.done() and worker not in self.cancelled_workers
        ]

    @property
    def finished_workers(self) -> list[Future[Any]]:
        """Returns a list of completed workers."""
        return [worker for worker in self.workers if worker.done()]

    @property
    def num_cancelled_workers(self) -> int:
        return len(self.cancelled_workers)

    @property
    def num_completed_workers(self) -> int:
        return len(self.completed_workers)

    @property
    def num_finished_workers(self) -> int:
        return len(self.finished_workers)

    @property
    def num_running_workers(self) -> int:
        return len(self.running_workers)

    @property
    def num_unfinished_workers(self) -> int:
        return len(self.unfinished_workers)

    @property
    def num_workers(self) -> int:
        return len(self.workers)

    @property
    def open_slots(self) -> int:
        """Returns the difference between max_workers and number of running workers."""
        return self.max_workers - self.num_running_workers

    @property
    def running_workers(self) -> list[Future[Any]]:
        """Returns a list of currently executing futures."""
        return [worker for worker in self.workers if worker.running()]

    @property
    def unfinished_workers(self) -> list[Future[Any]]:
        """Returns a list of unfinished futures."""
        return [worker for worker in self.workers if not worker.done()]

    def add_future(self, future: Future[Any]):
        """Add a future to `self.workers`."""
        self.workers.append(future)

    def cancel_workers(self):
        """Attempt to cancel any unfinished futures."""
        for worker in self.unfinished_workers:
            worker.cancel()

    def shutdown(self):
        """Attempt to cancel remaining futures and wait for running futures to finish."""
        if self.num_finished_workers < self.num_workers:
            console.print(f"{color_map.o1}Attempting to cancel remaining workers...")
            self.cancel_workers()
        if running_workers := self.running_workers:
            console.print(
                f"{color_map.c}Waiting for {color_map.sg2}{len(running_workers)}[/] workers to finish..."
            )
            num_running: Callable[
                [list[Future[Any]]], str
            ] = lambda n: f"[pink1]{len(n)} running workers..."
            with Console().status(
                num_running(running_workers), spinner="arc", spinner_style="deep_pink1"
            ) as c:
                while running_workers := self.running_workers:
                    c.update(num_running(running_workers))


class UrlManager:
    """Manages crawled and uncrawled urls."""

    def __init__(self):
        self._crawled: deque[Url] = deque()
        self._uncrawled: deque[Url] = deque()
        # Separate lists for schemeless urls so we don't have to restrip the whole list
        # everytime we check if a url is already in the list
        self._schemeless: deque[Url] = deque()
        self._schemeless_crawled: deque[Url] = deque()

    @property
    def crawled(self) -> deque[Url]:
        """Urls that have been or are currently being crawled."""
        return self._crawled

    @property
    def total(self) -> int:
        """Total crawled and uncrawled urls."""
        return len(self._schemeless)

    @property
    def uncrawled(self) -> deque[Url]:
        """Urls that have yet to be crawled."""
        return self._uncrawled

    def add_urls(self, urls: Sequence[Url]):
        """Append `urls` to `self.uncrawled_urls`."""
        for url in urls:
            self._uncrawled.append(url)

    def filter_urls(self, urls: Sequence[Url]) -> deque[Url]:
        """
        Filters out:
          * duplicate urls
          * already crawled urls
          * urls with schemes other than `http` and `https`
        """
        filtered_urls: deque[Url] = deque()
        for url in set(urls):
            if not url.scheme.startswith("http"):
                continue
            # Prevents duplicates where only diff is http vs https
            schemeless_url = url.schemeless
            if schemeless_url not in self._schemeless:
                self._schemeless.append(schemeless_url)
                filtered_urls.append(url)
        return filtered_urls

    def get_uncrawled(self) -> Url | None:
        """Get an uncrawled url from the front of the list.

        Returns `None` if uncrawled list is empty."""
        while self._uncrawled:
            url = self._uncrawled.popleft()
            schemeless_url = url.schemeless
            # double check url hasn't been crawled (cause threading)
            if schemeless_url not in self._schemeless_crawled:
                self._schemeless_crawled.append(schemeless_url)
                self._crawled.append(url)
                return url
        return None


class CrawlLimit(abc.ABC):
    @property
    @abc.abstractmethod
    def exceeded(self) -> bool:
        """Return whether this limit has been exceeded or not."""

    def __str__(self) -> str:
        """Customize the message used when this limit is exceeded."""
        return "Scraper limits exceeded."

    def __rich__(self) -> str:
        return self.__str__()


class MaxDepthLimit(CrawlLimit):
    def __init__(self, max_depth: int | None, thread_manager: ThreadManager):
        self.max_depth = max_depth
        self.thread_manager = thread_manager

    @property
    @override
    def exceeded(self) -> bool:
        if self.max_depth:
            return self.thread_manager.num_completed_workers >= self.max_depth
        return False

    @property
    def should_block(self) -> bool:
        """Whether adding new workers should be blocked."""
        if self.max_depth:
            return self.thread_manager.num_workers >= self.max_depth
        return False

    def __str__(self) -> str:
        return f"Max depth of {self.max_depth} exceeded."

    def __rich__(self) -> str:
        return f"Max depth of [bright_red]{self.max_depth}[/] exceeded."


class MaxTimeLimit(CrawlLimit):
    def __init__(self, max_time: float | None, timer: Timer):
        self.max_time = max_time
        self.timer = timer

    @property
    @override
    def exceeded(self) -> bool:
        if self.max_time:
            return self.timer.elapsed > self.max_time
        return False

    def __str__(self) -> str:
        return f"Max time of {self.timer.format_time(self.max_time) if self.max_time else 'None'} exceeded."

    def __rich__(self) -> str:
        return f"Max time of [bright_red]{self.timer.format_time(self.max_time) if self.max_time else 'None'}[/] exceeded."


class LimitCheckerMixin:
    @staticmethod
    def get_limits(obj: Any) -> list[CrawlLimit]:
        """Returns a list of `CrawlLimit` instances from `obj`."""
        return [
            member
            for member in obj.__dict__.values()
            if issubclass(type(member), CrawlLimit)
        ]

    @property
    def limits(self) -> list[CrawlLimit]:
        """Returns a list of `CrawlLimit` objects belonging to this instance."""
        return self.get_limits(self)

    @property
    def limits_exceeded(self) -> bool:
        return any(limit.exceeded for limit in self.limits)

    @property
    def exceeded_limits(self) -> list[CrawlLimit]:
        return [limit for limit in self.limits if limit.exceeded]


class CrawlScraper(ParserMixin, ScraperMetricsMixin, loggi.LoggerMixin, ChoresMixin):
    def __init__(self):
        super().__init__()

    @override
    def flush_items(self):
        self.parsable_items: deque[Any] = deque()  # type: ignore
        self.parsed_items: deque[Any] = deque()  # type: ignore

    @override
    def parse_item_wrapper(self, item: Any) -> Any:
        """Returns a parsed item or `None` if parsing failed."""
        try:
            parsed_item = self.parse_item(item)
            self.success_count += 1
            return parsed_item
        except Exception as e:
            self.logger.exception("Failure to parse item:")
            self.logger.error(str(item))
            self.fail_count += 1
            return None

    def scrape(self, source: Response):
        """
        Scrape `source` and store results.
        """
        try:
            parsable_items = self.get_parsable_items(source)
            self.logger.info(
                f"Returned {len(parsable_items)} parsable items from `{source.url}`."
            )
        except Exception as e:
            self.logger.exception(f"Error getting parsable items from `{source.url}`.")
        else:
            # Since likely usage is multithreaded,
            # don't want to pull from shared `self.parsed_items` container
            # when storing items
            parsed_items = self.parse_items(parsable_items)
            for item in parsed_items:
                self.parsed_items.append(item)
            self.store_items(parsed_items)

    @abc.abstractmethod
    def store_items(self, items: Sequence[Any]) -> None:
        """
        Unless the crawler using this scraper has `max_threads` set to 1,
        storage method/destination should be thread safe.

        Override this method with `pass` or `...` if you'd rather use a different function
        to store all parsed items after the crawl is completed.
        """


class Crawler(loggi.LoggerMixin, ChoresMixin, LimitCheckerMixin):
    """A mutli-threaded web crawler framework."""

    def __init__(
        self,
        scrapers: Sequence[CrawlScraper] = [],
        max_depth: int | None = None,
        max_time: float | None = None,
        log_name: str | int | loggi.LogName = loggi.LogName.CLASSNAME,
        log_dir: Pathish = "logs",
        max_threads: int = 3,
        same_site_only: bool = True,
        custom_url_manager: UrlManager | None = None,
    ):
        """
        Create a `Crawler` instance.

        #### :params:
        * `scraper`: An optional list of `CrawlScraper` or child instances to implement what happens for each crawled page.
        (NOTE: The `scraper` instances will have its logger set to this crawler's logger.)
        * `max_depth`: The maximum number of pages to crawl.
        * `max_time`: The maximum amount of time to crawl in seconds.
        * `log_name`: The file stem for the log file. Defaults to this instance's class name.
        * `log_dir`: The directory to write the log file to. Defaults to "logs".
        * `max_threads`: The max number of threads to use.
        * `same_site_only`: When `True`, only urls pointing to the same website will be added to the crawl queue.
        * `custom_url_manager`: An optional instance that inherits from `gruel.UrlManager`.
        """
        self.init_logger(log_name, log_dir)
        self.url_manager = custom_url_manager or UrlManager()
        self.thread_manager = ThreadManager(max_threads)
        self.timer = Timer()
        self.max_time = MaxTimeLimit(max_time, self.timer)
        self.max_depth = MaxDepthLimit(max_depth, self.thread_manager)
        self.same_site_only = same_site_only
        self._scrapers: list[CrawlScraper] = []
        self._was_cancelled = False
        for scraper in scrapers:
            self.register_scraper(scraper)

    @property
    def display_columns(self) -> list[ProgressColumn]:
        """The display columns to be used by the progress bar."""
        columns = list(Progress.get_default_columns())
        columns[3] = TimerColumn(True)
        return columns

    @property
    def finished(self) -> bool:
        """Returns `True` if there are no uncrawled urls and no unfinished threads."""
        return not (
            self.url_manager.uncrawled or self.thread_manager.unfinished_workers
        )

    @property
    @override
    def limits(self) -> list[CrawlLimit]:
        limits = super().limits
        # ? Should scraper limits only halt that particular scraper or the whole crawl
        # ? Separate class or flag to have both options
        for scraper in self.scrapers:
            limits.extend(self.get_limits(scraper))
        return limits

    @property
    def scrapers(self) -> list[CrawlScraper]:
        """The scraper being used by this crawler."""
        return self._scrapers

    def register_scraper(self, scraper: CrawlScraper):
        """
        Add this `scraper` instance to the list of scrapers.

        This `Crawler` instance's logger will be passed to the logger attribute of `scraper`.
        """
        scraper.logger = self.logger
        self._scrapers.append(scraper)

    @property
    def starting_url(self) -> Url:
        """The starting url of the last crawl."""
        return self._starting_url

    @property
    def was_cancelled(self) -> bool:
        """Returns if this crawler was intentionally terminated (e.g. keyboard interrupt)."""
        return self._was_cancelled

    def _dispatch_workers(self, executor: ThreadPoolExecutor):
        """Dispatch workers if there are open slots and new urls to be scraped."""
        while self.thread_manager.open_slots and not self.max_depth.should_block:
            url = self.url_manager.get_uncrawled()
            if url:
                self.thread_manager.add_future(executor.submit(self._handle_page, url))
            else:
                break

    def _handle_page(self, url: Url):
        self.logger.info(f"Scraping `{url}`.")
        response = self.request_page(url)
        urls = self.extract_crawlable_urls(response.get_linkscraper())
        new_urls = self.url_manager.filter_urls(urls)
        self.logger.info(f"Found {len(new_urls)} new urls on `{url}`.")
        self.url_manager.add_urls(new_urls)
        for scraper in self.scrapers:
            scraper.scrape(response)

    def print_exceeded_limits(self):
        for limit in self.exceeded_limits:
            self.logger.info(str(limit))
            console.print(limit)

    def _update_progress(self, progress: Progress, task: TaskID):
        num_completed = self.thread_manager.num_completed_workers
        total = self.thread_manager.num_workers + len(self.url_manager.uncrawled)
        progress.update(
            task,
            total=total,
            completed=num_completed,
            description=f"{num_completed}/{total} urls",
        )

    def crawl(self, starting_url: str):
        """Start crawling at `starting_url`."""
        self._starting_url = Url(starting_url)
        self.url_manager.add_urls([self._starting_url])
        self.prescrape_chores()
        with ThreadPoolExecutor(self.thread_manager.max_workers) as executor:
            try:
                with Progress(*self.display_columns) as progress:
                    crawler = progress.add_task()
                    while not self.finished and not self.limits_exceeded:
                        self._dispatch_workers(executor)
                        self._update_progress(progress, crawler)
                        time.sleep(0.1)
                    self._update_progress(progress, crawler)
                self.print_exceeded_limits()
            except KeyboardInterrupt:
                self._was_cancelled = True
            except Exception as e:
                print(str(e))
                self.thread_manager.shutdown()
                self.logger.close()
                raise e
            self.thread_manager.shutdown()
        self.postscrape_chores()
        self.logger.close()

    def extract_crawlable_urls(self, linkscraper: scrapetools.LinkScraper) -> list[Url]:
        """Returns a list of urls that can be added to the crawl list."""
        urls = [
            Url(link).fragmentless
            for link in linkscraper.get_links(
                "page",
                excluded_links=linkscraper.get_links("img"),
                same_site_only=self.same_site_only,
            )
        ]
        if self.same_site_only:
            urls = [url for url in urls if self.starting_url.is_same_site(url)]
        return urls

    @override
    def postscrape_chores(self):
        self.timer.stop()
        self.logger.info(f"Crawl completed in {self.timer.elapsed_str}.")
        console.print(
            f"{color_map.sg3}Crawl completed in {color_map.go1}{self.timer.elapsed_str}[/]."
        )
        for scraper in self.scrapers:
            scraper.postscrape_chores()

    @override
    def prescrape_chores(self):
        self.timer.start()
        for scraper in self.scrapers:
            scraper.prescrape_chores()
        start_time = f"{datetime.now():%H:%M:%S}"
        self.logger.info(f"Starting crawl ({start_time}) at {self.starting_url}")
        console.print(
            f"Starting crawl ({color_map.go1}{start_time}[/]) at {color_map.or_}{self.starting_url}"
        )

    def request_page(self, url: Url) -> Response:
        """Make a request to `url` and return the page."""
        return request(url.address, logger=self.logger)


class SeleniumCrawler(Crawler):
    """
    Requires Firefox to be installed and for `geckodriver.exe` to be in system PATH.

    Currently `max_threads` is hardcoded to `1`.
    """

    @override
    def __init__(
        self,
        scrapers: Sequence[CrawlScraper] = [],
        max_depth: int | None = None,
        max_time: float | None = None,
        log_name: str | int | loggi.LogName = loggi.LogName.CLASSNAME,
        log_dir: Pathish = "logs",
        same_site_only: bool = True,
        custom_url_manager: UrlManager | None = None,
        headless: bool = True,
        download_dir: Pathish | None = None,
    ):
        super().__init__(
            scrapers,
            max_depth,
            max_time,
            log_name,
            log_dir,
            1,
            same_site_only,
            custom_url_manager,
        )
        if download_dir:
            download_dir = Pathier(download_dir)
        self.user = User(headless, download_dir=Pathier(download_dir) if download_dir else None)  # type: ignore

    @override
    def request_page(self, url: Url) -> SeleniumResponse:
        self.user.get(url.address)
        return SeleniumResponse.from_selenium_user(self.user)

    @override
    def crawl(self, starting_url: str):
        super().crawl(starting_url)
        self.user.close_browser()



import logging
from typing import Any

import loggi
import requests
import requests.adapters
import requests.cookies
import scrapetools
import urllib3.util  # type: ignore
from bs4 import BeautifulSoup
from noiftimer import Timer
from seleniumuser.seleniumuser import User
from typing_extensions import Self, override
from whosyouragent import whosyouragent


class Response(requests.Response):
    """
    Override of `requests.Response` adding the following convenience methods:

    * `get_soup()`
    * `get_linkscraper()`
    """

    def get_soup(self, features: str = "html.parser") -> BeautifulSoup:
        """Returns a `BeautifulSoup` instance for this response."""
        return BeautifulSoup(self.text, features)

    def get_linkscraper(self) -> scrapetools.LinkScraper:
        """Returns a `LinkScraper` object from a `Response`."""
        linkscraper = scrapetools.LinkScraper(self.text, self.url)
        linkscraper.scrape_page()
        return linkscraper

    @classmethod
    def from_base_response(cls, response: requests.Response) -> Self:
        """Convert a `requests.Response` object into a `gruel.Response` object."""
        self = cls()
        self.__dict__ = response.__dict__.copy()
        return self


class SeleniumResponse(Response):
    """
    For mocking a `Response` type object from a `User` instance.

    Only mocks `url` and `text` variables/properties.
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._text = ""

    @property
    @override
    def text(self) -> str:
        return self._text

    @classmethod
    def from_selenium_user(cls, user: User) -> Self:
        """
        Construct a `Response` object from a `seleniumuser.User` instance.

        Pass the `User` instance to this function after using the instance to request a page.
        """
        self = cls()
        self.url = user.current_url()
        assert user.browser
        self._text = user.browser.page_source
        return self


retry_on_codes = [408, 413, 444, 499, 500, 502, 503, 504]


class Session(requests.Session):
    @override
    def __init__(
        self,
        randomize_useragent: bool = True,
        clear_cookies: bool = True,
        retry_count: int = 3,
        retry_backoff_factor: float = 0.1,
        retry_on_codes: list[int] = retry_on_codes,
        logger: loggi.Logger | logging.Logger | None = None,
    ):
        """
        Create a `Session` object.

        #### :params:
        `randomize_useragent`: If `True`, each request will have a randomized `User-Agent` string.
        `clear_cookies`: If `True`, cookies will be cleared from the session prior to each request.
        `retry_count`: The number of times to retry a failed request.
        `retry_backoff_factor`: For each failed request, the time before retrying will be `retry_backoff_factor * (2 ** retry_number)`
        `retry_on_codes`: List of status codes to retry requests on. Default is `[408, 413, 444, 499, 500, 502, 503, 504]`.
        `logger`: A logging instance to use.
        """
        super().__init__()
        self.randomize_useragent = randomize_useragent
        self.clear_cookies = clear_cookies
        self.timeout = 10
        self.set_retry(
            total=retry_count,
            backoff_factor=retry_backoff_factor,
            status_forcelist=retry_on_codes,
        )
        self.logger = logger

    def set_retry(self, *args: Any, **kwargs: Any):
        """Set the retry policy for failed requests.

        `*args` and `**kwargs` are any parameters accepted by `urllib3.util.Retry()`."""
        retries = urllib3.util.Retry(*args, **kwargs)
        self.mount("http://", requests.adapters.HTTPAdapter(max_retries=retries))
        self.mount("https://", requests.adapters.HTTPAdapter(max_retries=retries))

    @override
    def prepare_request(self, request: requests.Request) -> requests.PreparedRequest:
        if self.randomize_useragent:
            self.headers["User-Agent"] = whosyouragent.get_agent()
        if self.clear_cookies:
            self.cookies = requests.cookies.RequestsCookieJar()
        return super().prepare_request(request)

    @override
    def request(self, *args: Any, **kwargs: Any) -> Response:
        response = super().request(*args, **kwargs)
        return Response.from_base_response(response)

    @override
    def send(
        self, request: requests.PreparedRequest, **kwargs: Any
    ) -> requests.Response:
        if self.logger:
            self.logger.info(
                f"Sending a `{request.method}` request to `{request.url}`."
            )
        try:
            response = super().send(request, **kwargs)
        except Exception as e:
            if self.logger:
                self.logger.exception(
                    f"`{request.method}` request to `{request.url}` failed."
                )
            raise e
        if self.logger:
            logged_response = response
            # Without checking `history` all redirected responses will be logged with the same details as the final response
            for response_ in response.history:
                if response_.url == request.url:
                    logged_response = response_
                    break
            self.logger.info(
                f"Request to `{request.url}` completed with status code `{logged_response.status_code}` in {Timer.format_time(logged_response.elapsed.total_seconds(), True)}."
            )
        return response


def request(
    url: str,
    method: str = "get",
    retry_count: int = 3,
    retry_backoff_factor: float = 0.1,
    retry_on_codes: list[int] = retry_on_codes,
    logger: loggi.Logger | logging.Logger | None = None,
    *args: Any,
    **kwargs: Any,
) -> Response:
    """
    Constructs and sends a :class:`Request <Request>`.

    `url`: URL for the new :class:`Request` object.
    `method`: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.

    * `retry_count`: The number of times to retry a failed request.
    * `retry_backoff_factor`: For each failed request, the time before retrying will be `retry_backoff_factor * (2 ** retry_number)`
    * `logger`: A logging instance to use.
    * `retry_on_codes`: List of status codes to retry requests on. Default is `[408, 413, 444, 499, 500, 502, 503, 504]`.

    `params`: dict, list of tuples or bytes to send in the query string for the :class:`Request`.
    `data`: dict, list of tuples, bytes, or file-like object to send in the body of the :class:`Request`.
    `json`: A JSON serializable Python object to send in the body of the :class:`Request`.
    `headers`: dict of HTTP Headers to send with the :class:`Request`. The `User-Agent` header will be randomized unless supplied.
    `cookies`: dict or CookieJar object to send with the :class:`Request`.
    `files`: dict of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
        ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
        or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
        defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
        to add for the file.
    `auth`: Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    `timeout`: float | tuple, How many seconds to wait for the server to send data
        before giving up, as a float, or a :ref:`(connect timeout, read
        timeout) <timeouts>` tuple.
    `allow_redirects`: bool. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
    `proxies`: dict mapping protocol to the URL of the proxy.
    `verify`: Either a bool, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
    `stream`: if ``False``, the response content will be immediately downloaded.
    `cert`: if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
    """
    with Session(
        retry_count=retry_count,
        retry_backoff_factor=retry_backoff_factor,
        logger=logger,
        retry_on_codes=retry_on_codes,
    ) as session:
        return session.request(method, url, *args, **kwargs)



from typing import Any

from typing_extensions import Sequence, override

from gruel import Gruel


class SubGruel(Gruel):
    @override
    def get_source(self) -> Any:
        raise NotImplementedError

    @override
    def get_parsable_items(self, source: Any) -> list[Any]:
        raise NotImplementedError

    @override
    def parse_item(self, item: Any) -> Any:
        raise NotImplementedError

    @override
    def store_items(self, items: Sequence[Any]) -> None:
        raise NotImplementedError



import argparse
import importlib
import importlib.machinery
import importlib.util
import inspect
from types import ModuleType
from typing import Any, Sequence, Type

import loggi
import quickpool
from pathier import Pathier, Pathish
from younotyou import Matcher, younotyou

from .core import ChoresMixin, Gruel


class GruelFinder(loggi.LoggerMixin):
    """Find and load classes that subclass `Gruel`."""

    def __init__(
        self,
        subgruel_classes: list[str] = ["*"],
        file_exclude_patterns: list[str] = [],
        scan_path: Pathier | None = None,
        file_include_patterns: list[str] = ["*.py"],
        recursive: bool = True,
        log_dir: Pathish = "logs",
    ):
        """#### :params:

        `subgruel_classes`: A list of class names for scrapers that should be loaded.
        In order to be loaded, a scraper class must have a name in this list and have `Gruel` somewhere in its inheritance hierarchy.
        Can use wildcard ('*') patterns for matching.

        `file_exclude_patterns`: Files that match these patterns will not be scanned.

        `scan_path`: The path to scan for scraper classes.

        `file_include_patterns`: Files that match these patterns will be scanned.

        `recursive`: Whether the scan should be recursive or not.

        `log_dir`: The directory this instance's log should be saved to.

        Will find and load all classes in the "scrapers" directory that inherit from `Gruel`
        and start with "MySubGruel", but don't contain "Scratch" in the name:
        >>> finder = finder(["MySubGruel*"], ["*Scratch*"], "scrapers")
        >>> gruels = finder.find()"""
        self.subgruel_classes = subgruel_classes
        self.file_exclude_patterns = file_exclude_patterns
        self.scan_path = scan_path or Pathier.cwd()
        self.file_include_patterns = file_include_patterns
        self.recursive = recursive
        self.init_logger("gruel_finder", log_dir)

    def get_bases(self, object: Any) -> list[Any]:
        """Returns a recursive list of all the classes `object` inherits from."""
        parents: list[Any] = []
        bases = object.__bases__
        if not bases:
            return parents
        for base in bases:
            parents.append(base)
            parents.extend(self.get_bases(base))
        return parents

    def is_subgruel(self, object: Any) -> bool:
        """Returns whether `object` inherits from `Gruel` somewhere in its ancestory."""
        if not inspect.isclass(object) or Gruel not in self.get_bases(object):
            return False
        return True

    def glob_files(self) -> list[Pathier]:
        """Search `self.scan_path` for files according to `self.file_include_patterns` and `self.file_exclude_patterns`.

        Returns the file list."""
        globber = self.scan_path.rglob if self.recursive else self.scan_path.glob
        files = [
            str(file)
            for pattern in self.file_include_patterns
            for file in globber(pattern)
        ]
        files = [
            Pathier(file)
            for file in younotyou(files, exclude_patterns=self.file_exclude_patterns)
        ]
        return files

    def load_module_from_file(self, file: Pathier) -> ModuleType | None:
        """Attempts to load and return a module from `file`."""
        module_name = file.stem
        try:
            module = importlib.machinery.SourceFileLoader(
                module_name, str(file)
            ).load_module()
            self.logger.info(f"Successfully imported `{module_name}` from `{file}`.")
            return module
        except Exception as e:
            self.logger.exception(f"Failed to load `{module_name}` from `{file}`.")

    def strain_for_gruel(self, modules: list[ModuleType]) -> list[Type[Gruel]]:
        """Searches `modules` for classes that inherit from `Gruel` and are in `self.subgruel_classes`.

        Returns the list of classes."""
        matcher = Matcher(self.subgruel_classes)
        return [
            getattr(module, class_)
            for module in modules
            for class_ in dir(module)
            if class_ in matcher and self.is_subgruel(getattr(module, class_))
        ]

    def find(self) -> list[Type[Gruel]]:
        """Run the scan and return `Gruel` subclasses."""
        files = self.glob_files()
        modules: list[ModuleType] = []
        for file in files:
            if module := self.load_module_from_file(file):
                modules.append(module)
        return self.strain_for_gruel(modules)


class Brewer(loggi.LoggerMixin, ChoresMixin):
    """Use to do multithreaded execution of a list of scrapers.

    Intended to be used with `Gruel` scrapers, but anything with a `scrape` method can be passed.

    To run any `Gruel` scrapers from the current directory:
    >>> Brewer(GruelFinder().find()).brew()

    The `prescrape_chores` and `postscrape_chores` can be set/overridden like the same methods in `Gruel`.

    When calling the `brew` method they will be executed once before and after all the scrapers have been executed.

    i.e.
    >>> brewer = Brewer(GruelFinder().find())
    >>> brewer.prescrape_chores()
    >>> results = brewer.scrape()
    >>> brewer.postscrape_chores()

    is equivalent to
    >>> results = Brewer(GruelFinder().find()).brew()

    except `brew()` has some logging."""

    def __init__(
        self,
        scrapers: Sequence[Type[Gruel]],
        scraper_args: Sequence[Sequence[Any]] = [],
        scraper_kwargs: Sequence[dict[str, Any]] = [],
        log_dir: Pathish = "logs",
    ):
        """#### :params:

        `scrapers`: A list of scraper classes to initialize and execute.
        A scraper should not be instantiated before being passed.
        When `Brewer` runs a scraper it will instantiate the object at execution time and call it's `scrape` method.

        `scraper_args`: A list where each element is a list of positional arguments to be passed to the corresponding scraper's `__init__` function.

        `scraper_kwargs`: A list of dictionaries where each dictionary is a set of keyword arguments to be passed to the corresponding scraper's `__init__` function.

        `log_dir`: The directory to store `Brewer` logs in. Defaults to "logs".

        e.g.
        >>> class MyGruel(Gruel):
        >>>   def __init__(self, value:int):
        >>>     super().__init__()
        >>>     self.value = value
        >>>
        >>>   def scrape(self)->int:
        >>>     return self.value
        >>>
        >>> num_scrapers = 5
        >>> values = list(range(5))
        >>> brewer = Brewer(
        >>>   [MyGruel]*num_scrapers,
        >>>   [(val,) for val in values]
        >>> results = brewer.brew()
        >>> print(results)
        >>> [0, 1, 2, 3, 4]"""
        self.init_logger(log_dir=log_dir)
        self.scrapers = scrapers
        num_scrapers = len(self.scrapers)
        # Pad args and kwargs if there aren't any given
        self.scraper_args: Sequence[Any] = scraper_args or [[]] * num_scrapers
        self.scraper_kwargs: Sequence[dict[str, Any]] = (
            scraper_kwargs or [{}] * num_scrapers
        )

    def _prep_scrapers(self) -> list[tuple[Any, Sequence[Any], dict[str, Any]]]:
        return [
            (scraper, args, kwargs)
            for scraper, args, kwargs in zip(
                self.scrapers, self.scraper_args, self.scraper_kwargs
            )
        ]

    def scrape(self) -> list[Any]:
        """Run the `scrape()` method for each scraper in `scrapers`.

        Execution is multithreaded."""

        def execute(scraper: Type[Gruel], args: Sequence[Any], kwargs: dict[str, Any]):
            return scraper(*args, **kwargs).scrape()

        pool = quickpool.ThreadPool(
            [execute] * len(self.scrapers), self._prep_scrapers()
        )
        return pool.execute()

    def brew(self) -> list[Any] | None:
        """Execute pipeline.

        1. self.prescrape_chores()
        2. self.scrape()
        3. self.postscrape_chores()"""

        try:
            self.logger.logprint("Beginning brew")
            # 1--------------------------------------------
            self.logger.logprint("Executing prescrape chores")
            self.prescrape_chores()
            # 2--------------------------------------------
            self.logger.logprint("Starting scrape")
            results = self.scrape()
            self.logger.logprint("Scrape complete")
            # 4--------------------------------------------
            self.logger.logprint("Executing postscrape chores")
            self.postscrape_chores()
            self.logger.logprint("Brew complete")
            return results
        except Exception as e:
            print(e)
            self.logger.exception("Exception occured during brew():")


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="Brewer", description="Invoke `Brewer` from the command line."
    )

    parser.add_argument(
        "subgruel_classes",
        type=str,
        nargs="*",
        help=""" A list of Gruel scraper class names to find and import. """,
    )
    parser.add_argument(
        "-e",
        "--excludes",
        type=str,
        nargs="*",
        default=[],
        help=""" A list of glob style file patterns to exclude from the scan. """,
    )
    parser.add_argument(
        "-i",
        "--includes",
        type=str,
        nargs="*",
        default=["*.py"],
        help=""" A list of glob style file patterns to include in the scan. Defaults to "*.py". """,
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default=Pathier.cwd(),
        help=""" The directory path to scan. Defaults to the current working directory. """,
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help=""" Whether -p/--path should be scanned recursively or not. """,
    )
    parser.add_argument(
        "-l",
        "--log_dir",
        type=str,
        default="logs",
        help=""" The directory to save the brew log to.""",
    )
    args = parser.parse_args()
    args.path = Pathier(args.path)

    return args


def main(args: argparse.Namespace | None = None):
    if not args:
        args = get_args()
    finder = GruelFinder(
        args.subgruel_classes,
        args.excludes,
        args.path,
        args.includes,
        args.recursive,
        args.log_dir,
    )
    brewer = Brewer(
        finder.find(),
        args.log_dir,
    )
    brewer.brew()


if __name__ == "__main__":
    main(get_args())



from dataclasses import dataclass
from urllib.parse import SplitResult, urlsplit, urlunsplit

from typing_extensions import Self


@dataclass
class Url:
    address: str

    def __post_init__(self):
        self.address = self.address.strip("/ ")

    @property
    def _parsed(self) -> SplitResult:
        """Returns this instance's `address` split into scheme, netloc, path, query, and fragment."""
        return urlsplit(self.address.strip("/ "))

    def _from_parts(
        self, scheme: str, netloc: str, path: str, query: str, fragment: str
    ) -> Self:
        """Return a new `Url` object from url parts."""
        return self.__class__(urlunsplit((scheme, netloc, path, query, fragment)))

    @property
    def scheme(self) -> str:
        """The scheme of the url."""
        return self._parsed.scheme

    @scheme.setter
    def scheme(self, scheme: str):
        self.address = self._from_parts(
            scheme, self.netloc, self.path, self.query, self.fragment
        ).address

    @property
    def netloc(self) -> str:
        """The netloc (net location) of the url."""
        return self._parsed.netloc

    @netloc.setter
    def netloc(self, netloc: str):
        self.address = self._from_parts(
            self.scheme, netloc, self.path, self.query, self.fragment
        ).address

    @property
    def path(self) -> str:
        """The path portion of the url (contains leading slash)."""
        return self._parsed.path

    @path.setter
    def path(self, path: str):
        self.address = self._from_parts(
            self.scheme, self.netloc, path, self.query, self.fragment
        ).address

    @property
    def query(self) -> str:
        """The query portion of the url."""
        return self._parsed.query

    @query.setter
    def query(self, query: str):
        self.address = self._from_parts(
            self.scheme, self.netloc, self.path, query, self.fragment
        ).address

    @property
    def fragment(self) -> str:
        """The fragment portion of the url."""
        return self._parsed.fragment

    @fragment.setter
    def fragment(self, fragment: str):
        self.address = self._from_parts(
            self.scheme, self.netloc, self.path, self.query, fragment
        ).address

    @property
    def fragmentless(self) -> Self:
        """A `Url` object with no fragment."""
        return self._from_parts(self.scheme, self.netloc, self.path, self.query, "")

    @property
    def base(self) -> Self:
        """Returns a `Url` with just scheme and netloc."""
        return self._from_parts(self.scheme, self.netloc, "", "", "")

    @property
    def schemeless(self) -> Self:
        """Returns a `Url` with no scheme."""
        return self._from_parts("", self.netloc, self.path, self.query, self.fragment)

    def is_same_site(self, other: Self) -> bool:
        """Returns `True` if `other` appears to be the same website/host."""
        netloc1 = self.netloc.removeprefix("www.")
        netloc2 = other.netloc.removeprefix("www.")
        # Account for one being a subdomain of the other
        if len(netloc1) < len(netloc2):
            return netloc1 in netloc2
        return netloc2 in netloc1

    def __str__(self) -> str:
        return self.address

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__) or not issubclass(
            type(other), self.__class__
        ):
            raise ValueError(f"Can't compare `Url` object to non-`Url` object.")
        return self.address == getattr(other, "address")

    def __hash__(self) -> int:
        return hash(self.address)



import abc
import inspect

import loggi
from noiftimer import Timer
from pathier import Pathier, Pathish
from printbuddies import track
from typing_extensions import Any, Sequence, override

from .requests import Response, request


class ChoresMixin:
    """Adds `prescrape_chores` and `postscrape_chores` methods to inheriting classes."""

    def postscrape_chores(self):
        """Chores to do after scraping."""
        ...

    # ? def interscrape_chores(self):

    def prescrape_chores(self):
        """Chores to do before scraping."""
        ...


class ParserMixin(abc.ABC):
    """Core parser functionality for scraper classes."""

    def __init__(self):
        super().__init__()
        self.flush_items()
        self._show_parse_items_prog_bar = False

    @property
    def show_parse_items_prog_bar(self) -> bool:
        return self._show_parse_items_prog_bar

    @show_parse_items_prog_bar.setter
    def show_parse_items_prog_bar(self, should_show: bool):
        self._show_parse_items_prog_bar = should_show

    def flush_items(self):
        """Flush `parsable_items` and `parsed_items`."""
        self.parsable_items: list[Any] = []
        self.parsed_items: list[Any] = []

    @abc.abstractmethod
    def get_parsable_items(self, source: Any) -> list[Any]:
        """Get atomic chunks to be parsed from `source` and return as a list."""

    @abc.abstractmethod
    def parse_item(self, item: Any) -> Any:
        """Parse `item` and return parsed data."""

    def parse_item_wrapper(self, item: Any) -> Any:
        """
        Override this to control what happens around `self.parse_item()` for each item (error handling etc.).

        This way related subclasses can have the same auxillary things happen on calls to `parse_item`
        while having different `parse_item` implementations.

        This method will be called by `parse_items` for each item.

        When overriding, you should call `parse_item`, pass it `item`, and return the result.

        Without overriding, this function simply calls and returns `self.parse_item(item)`.

        basic e.g.:
        >>> def parse_item_wrapper(self, item:Any)->Any:
        >>>   try:
        >>>     return self.parse_item(item)
        >>>   except Exception as e:
        >>>     print(e)
        """
        return self.parse_item(item)

    def parse_items(self, parsable_items: Sequence[Any]) -> list[Any]:
        """Parse items and return them."""
        parsed_items: list[Any] = []
        for item in track(parsable_items, disable=not self.show_parse_items_prog_bar):
            parsed_item = self.parse_item_wrapper(item)
            parsed_items.append(parsed_item)
        return parsed_items


class ScraperMetricsMixin:
    """Mixin for various run time scraper stats."""

    def __init__(self):
        super().__init__()
        self.timer: Timer = Timer()
        self.success_count: int = 0
        self.fail_count: int = 0
        self.failed_to_get_parsable_items: bool = False
        self.unexpected_failure_occured: bool = False

    @property
    def had_failures(self) -> bool:
        """`True` if getting parsable items, parsing items, or unexpected failures occured."""
        return (
            self.fail_count > 0
            or self.failed_to_get_parsable_items
            or self.unexpected_failure_occured
        )


class Gruel(ParserMixin, ScraperMetricsMixin, loggi.LoggerMixin, ChoresMixin):
    """
    Primary base class for building scrapers.

    Inheriting classes must implement the following abstract methods:
        * `def get_source(self) -> Any` (Retrieve source material to be parsed, typically a request response)
        * `def get_parsable_items(self, source: Any) -> list[Any]` (Return a list of parsable sub-chunks from `source`, e.g. days from a calendar page)
        * `def parse_item(self, item: Any) -> Any` (Extract and return desired data from sub-chunk, e.g. populating and returning a model from a `BeautifulSoup` object)
        * `def store_items(self, items: Sequence[Any]) -> None` (How parsed items should be saved, e.g. storing parsed items in a database)

    Optionally, inheriting classes may override:
        * `def prescrape_chores(self)` (Anything that should be done prior to scraping)
        * `def postscrape_chores(self)` (Anything that should be done after scraping)
        * `def parse_item_wrapper(self, item: Any) -> Any`

        Default behavior of `parse_item_wrapper` is to call `self.parse_item(item)`.

        If no exception is raised, `self.success_count` is incremented and the parsed item is returned.

        If an exception occurs, it is logged, `self.fail_count` is incremented, and `None` is returned.

        NOTE: Returning `None` means `self.parsable_items` and `self.parsed_items` will be equal length
        and `self.store_items()` should handle possible `None` arguments.
    """

    def __init__(
        self,
        name: str | None = None,
        log_dir: Pathish = "logs",
        show_parse_items_prog_bar: bool = False,
    ):
        """
        :params:
        * `name`: The name of this scraper. If `None`, the name will be the stem of the file this class/subclass was defined in.
        i.e. A `Gruel` subclass located in a file called `myscraper.py` will have the name `"myscraper"`.
        * `log_dir`: The directory this scraper's logs should be saved to.
        """
        super().__init__()
        self._name = name
        self.init_logger(self.name, log_dir)
        self.show_parse_items_prog_bar = show_parse_items_prog_bar

    @property
    def name(self) -> str:
        """Returns the name given to __init__ or the stem of the file this instance was defined in if one wasn't given."""
        return self._name or Pathier(inspect.getsourcefile(type(self))).stem  # type: ignore

    def _fetch_and_parse(self):
        """Fetch source content and pass to parsing workflow."""
        try:
            source = self.get_source()
        except Exception as e:
            self.logger.exception(f"Error getting source data.")
        else:
            self._parse_source(source)

    def _parse_source(self, source: Any):
        """
        Run the parsing workflow and handle errors.
        """
        try:
            self.parsable_items = self.get_parsable_items(source)
            self.logger.info(
                f"{self.name}:get_parsable_items() returned {len(self.parsable_items)} items."
            )
        except Exception:
            self.failed_to_get_parsable_items = True
            self.logger.exception(f"Error in {self.name}:get_parsable_items().")
        else:
            self.parsed_items = self.parse_items(self.parsable_items)
            self.logger.info(
                f"Scrape completed in {self.timer.elapsed_str} with {self.success_count} successes and {self.fail_count} failures."
            )

    @abc.abstractmethod
    def get_source(self) -> Any:
        """Should fetch and return the raw data to be scraped.

        Typically would request a webpage and return the response."""

    @override
    def parse_item_wrapper(self, item: Any) -> Any:
        """Returns a parsed item or `None` if parsing failed."""
        try:
            parsed_item = self.parse_item(item)
            self.success_count += 1
            return parsed_item
        except Exception as e:
            self.logger.exception("Failure to parse item:")
            self.logger.error(str(item))
            self.fail_count += 1
            return None

    def request(self, *args: Any, **kwargs: Any) -> Response:
        """
        Note: For convenience, passes this instances logger to the request functions

        Constructs and sends a :class:`Request <Request>`.

        `url`: URL for the new :class:`Request` object.
        `method`: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.

        * `retry_count`: The number of times to retry a failed request.
        * `retry_backoff_factor`: For each failed request, the time before retrying will be `retry_backoff_factor * (2 ** retry_number)`
        * `retry_on_codes`: List of status codes to retry requests on. Default is `[408, 413, 444, 499, 500, 502, 503, 504]`.

        `params`: dict, list of tuples or bytes to send in the query string for the :class:`Request`.
        `data`: dict, list of tuples, bytes, or file-like object to send in the body of the :class:`Request`.
        `json`: A JSON serializable Python object to send in the body of the :class:`Request`.
        `headers`: dict of HTTP Headers to send with the :class:`Request`. The `User-Agent` header will be randomized unless supplied.
        `cookies`: dict or CookieJar object to send with the :class:`Request`.
        `files`: dict of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
            ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
            or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content-type'`` is a string
            defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
            to add for the file.
        `auth`: Auth tuple to enable Basic/Digest/Custom HTTP Auth.
        `timeout`: float | tuple, How many seconds to wait for the server to send data
            before giving up, as a float, or a :ref:`(connect timeout, read
            timeout) <timeouts>` tuple.
        `allow_redirects`: bool. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
        `proxies`: dict mapping protocol to the URL of the proxy.
        `verify`: Either a bool, in which case it controls whether we verify
                the server's TLS certificate, or a string, in which case it must be a path
                to a CA bundle to use. Defaults to ``True``.
        `stream`: if ``False``, the response content will be immediately downloaded.
        `cert`: if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair.
        """
        kwargs["logger"] = self.logger
        return request(*args, **kwargs)

    def scrape(self):
        """Run the scraper:
        1. prescrape chores
        2. get parsable items
        3. parse items
        4. store items
        5. postscrape chores"""
        try:
            self.timer.start()
            self.logger.info("Scrape started.")
            self.prescrape_chores()
            self._fetch_and_parse()
            self.store_items(self.parsed_items)
        except Exception:
            self.unexpected_failure_occured = True
            self.logger.exception(f"Unexpected failure in {self.name}:scrape()")
        self.postscrape_chores()
        self.logger.close()

    @abc.abstractmethod
    def store_items(self, items: Sequence[Any]) -> None:
        """Store parsed items."""



from .brewer import Brewer, GruelFinder
from .core import ChoresMixin, Gruel, ParserMixin, ScraperMetricsMixin
from .crawler import (
    Crawler,
    CrawlLimit,
    CrawlScraper,
    LimitCheckerMixin,
    SeleniumCrawler,
    UrlManager,
)
from .models import Url
from .requests import Response, Session, request, retry_on_codes

__version__ = "4.4.3"
__all__ = [
    "Brewer",
    "GruelFinder",
    "Gruel",
    "ChoresMixin",
    "ParserMixin",
    "request",
    "Session",
    "Response",
    "retry_on_codes",
    "UrlManager",
    "Crawler",
    "CrawlScraper",
    "ScraperMetricsMixin",
    "CrawlLimit",
    "LimitCheckerMixin",
    "SeleniumCrawler",
    "Url",
]
